# scripts/migrage_creators_json_txt.py

import os
import json

from rcdl.core.models import Creator
from rcdl.core.config import Config
from rcdl.core.parser import get_domain, append_creator

JSON_PATH = Config.CACHE_DIR / "creators.json"

# check file exist
if not os.path.exists(JSON_PATH):
    print("creators.json deoes not exist. Check")

# load file
with open(JSON_PATH, "r") as f:
    json_creators = json.load(f)

# convert to Creator
creators = []
for json_creator in json_creators:
    creators.append(
        Creator(
            creator_id=json_creator["creator_id"],
            service=json_creator["service"],
            domain=get_domain(json_creator["service"]),
            status=None,
        )
    )

# save creator
for c in creators:
    append_creator(c)
    print(f"Saved new creator: {c.service}/{c.creator_id}")

print(f"You can now delete {JSON_PATH}")



# __main__.py

import logging

from rcdl.core.config import Config, setup_logging

# setup file structure
Config.ensure_dirs()
Config.ensure_files()

# setup logging
setup_logging(Config.LOG_FILE, level=0)

logging.info("--- INIT ---")
logging.info("Logger initialized")

# init database
from rcdl.core.db import DB  # noqa: E402

db = DB()
db.init_table()
logging.info(f"DB version: {db.get_schema_version()}")
db.close()

from rcdl.interface.cli import cli  # noqa: E402, F401



# core/fuse.py

import os

from rcdl.interface.ui import UI
from rcdl.core.db import DB
from rcdl.core.models import VideoStatus
from rcdl.core.config import Config
from rcdl.core.downloader_subprocess import ffmpeg_concat


def fuse_videos():
    """Fuse videos"""

    allowed_status = [
        VideoStatus.DOWNLOADED,
        VideoStatus.CONCAT_WIP,
        VideoStatus.CONCAT_FAILED,
        VideoStatus.CONCAT_DONE,  # dev only remove
    ]

    # load db videos
    with DB() as db:
        videos = db.query_videos(status=allowed_status, min_part_number=1)

    # get unique posts id
    posts_ids = set()
    for video in videos:
        posts_ids.add(video.post_id)

    with UI.progress_total_concat() as progress:
        task = progress.add_task("Total concat", total=len(posts_ids))

        for post_id in posts_ids:
            UI.info(f"Looking at post_id: {post_id}")

            # get all videos with same post_id
            with DB() as db:
                videos = db.query_videos(post_id=post_id)
            if not videos:
                UI.error("Query SQL Failed.")
                progress.update(task, advance=1)
                continue

            # check each videos of the same post is fully downloaded
            ok = True
            for video in videos:
                if video.status not in allowed_status:
                    ok = False
                    break
            if not ok:
                progress.update(task, advance=1)
                continue

            # sort by part number
            videos.sort(key=lambda v: int(v.part))

            ok = True
            for video in videos:
                # make sure video exist
                path = os.path.join(
                    Config.creator_folder(video.creator_id), video.relative_path
                )
                if not os.path.exists(path):
                    UI.error(f"Video @ {path} does not exists")
                    ok = False

                # if status is concat WIP, these should not be possible
                if video.status == VideoStatus.CONCAT_WIP:
                    UI.warning(
                        f"Video '{video.relative_path}' has status 'CONCAT_WIP'. This is a bug and should not be possible."
                    )

            # update videos status in db to CONCAT_WIP
            # in case of problems in the scripts, we will know
            with DB() as db:
                for video in videos:
                    db.set_status(video, VideoStatus.CONCAT_WIP)

            result = 1
            try:
                result = ffmpeg_concat(videos)
            except Exception as e:
                UI.error(f"Failed concat due to: {e}")

            # concat failed
            if not result == 0:
                with DB() as db:
                    for video in videos:
                        db.set_status(video, VideoStatus.CONCAT_FAILED)

            # concat succeeded
            with DB() as db:
                for video in videos:
                    db.set_status(video, VideoStatus.CONCAT_DONE)

            progress.update(
                task,
                advance=1,
                description=f"Concated videos for post id {post_id}",
            )



# core/downloader.py

import logging
import os

import requests

import rcdl.core.parser as parser
from .api import URL
from .config import Config
from .models import Creator, Video, VideoStatus
from .db import DB
from .downloader_subprocess import ytdlp_subprocess
from .file_io import write_json, load_json
from rcdl.interface.ui import UI


class PostsFetcher:
    """
    Fetch posts from api. Save as JSON. Handle multiple pages requests
    """

    def __init__(
        self, url: str, json_path: str, max_page: int = Config.DEFAULT_MAX_PAGE
    ):
        self.url = url
        self.json_path = json_path

        self.page = 0
        self.max_page = max_page

        self.status = 200

    def _request_page(self, url: str) -> requests.Response:
        """Request a single page and return json dict"""
        logging.info(f"RequestEngine url {url}")
        headers = URL.get_headers()
        response = requests.get(url, headers=headers)
        if response.status_code != 200:
            logging.warning(f"Failed request {url}: {requests.status_codes}")
        return response

    def request(self, params: dict = {}):
        with UI.progress_posts_fetcher(self.max_page) as progress:
            task = progress.add_task("Fetching posts", total=self.max_page)

            while self.status == 200 and self.page < self.max_page:
                o = self.page * Config.POST_PER_PAGE
                params["o"] = o
                url = URL.add_params(self.url, params)

                try:
                    # Dry run: not request acutally made
                    if Config.DRY_RUN:
                        logging.debug(
                            f"DRY-RUN posts fetcher {url} -> {self.json_path}"
                        )
                        self.page += 1
                        continue

                    response = self._request_page(url)
                    self.status = response.status_code

                    # if the programm crash while doing requests, previous requests are still saved and not overwritten.
                    if self.page > 0:
                        json_data = list(load_json(self.json_path))
                    else:
                        json_data = []

                    # for discover command, response json is in a different format and contains 'posts'
                    if self.status == 200:
                        if "posts" in response.json():
                            json_data.extend(response.json()["posts"])
                        else:
                            json_data.extend(response.json())

                        write_json(self.json_path, json_data, mode="w")

                    progress.update(
                        task,
                        advance=1,
                        description=f"Fetched {len(json_data)} posts (page {self.page + 1}/{self.max_page})",
                    )
                except Exception as e:
                    logging.error(f"Error in request {url} p{self.page}: {e}")
                finally:
                    self.page += 1


class VideoDownloader:
    """Handle downloading a list of Videos and update DB status"""

    def __init__(self):
        pass

    def _build_url(self, video: Video):
        return URL.get_url_from_file(video.domain, video.url)

    def _build_output_path(self, video: Video, discover: bool = False):
        if discover:
            return os.path.join(Config.DISCOVER_DIR, video.relative_path)

        return os.path.join(
            Config.creator_folder(video.creator_id), video.relative_path
        )

    def _update_db_status(self, result: int, video: Video):
        with DB() as d:
            if result == 0:
                d.set_status(video, VideoStatus.DOWNLOADED, fail_count=0)
            else:
                d.set_status(video, VideoStatus.FAILED, fail_count=video.fail_count + 1)

    def downloads(
        self, videos: list[Video], write_db: bool = True, discover: bool = False
    ):
        progress, task = UI.video_progress(total=len(videos))
        try:
            for video in videos:
                url = self._build_url(video)
                filepath = self._build_output_path(video, discover=discover)

                UI.set_current_video_progress(
                    f"{video.creator_id}@({video.service})", video.relative_path
                )

                if Config.DRY_RUN:
                    UI.debug(f"Dry run: dl {video.creator_id} @ {filepath}")
                    progress.advance(task)
                    continue

                if os.path.exists(filepath):
                    UI.warning(
                        f"Video {url} @ {filepath} already exists. Possible DB problem"
                    )
                    progress.advance(task)
                    continue

                result = ytdlp_subprocess(url, filepath)
                if write_db:
                    self._update_db_status(result, video)

                progress.advance(task)
        finally:
            UI.close_video_progress()


def fetch_posts_by_tag(tag: str, max_page: int = Config.DEFAULT_MAX_PAGE) -> dict:
    """Helper function to get all posts from a search results"""
    url = URL.get_posts_page_url_wo_param()
    path = Config.cache_file(tag)
    pf = PostsFetcher(url, str(path), max_page=max_page)
    pf.request(params={"tag": tag})

    return load_json(path)


def fetch_posts_by_creator(creator: Creator) -> dict:
    """Helper function to get all posts from a creator"""
    url = URL.get_creator_post_wo_param(creator)
    path = Config.cache_file(f"{creator.creator_id}_{creator.service}")
    pf = PostsFetcher(url, str(path))
    pf.request()

    return load_json(path)


def refresh_creators_videos():
    """
    Command refresh
    For each creator:
        - get all posts to a .json
        - from the .json filter to keep only the posts with videos in it
        - convert posts dict to Videos
        - update the DB
    """
    creators = parser.get_creators()
    for creator in creators:
        UI.info(f"Creator {creator.creator_id} from {creator.service}")

        fetch_posts_by_creator(creator)
        posts_with_videos = parser.filter_posts_with_videos_from_json(
            str(Config.cache_file(f"{creator.creator_id}_{creator.service}"))
        )
        all_videos = parser.convert_posts_to_videos(posts_with_videos)

        UI.info(
            f"Found {len(all_videos)} videos from {len(posts_with_videos)} posts with videos url"
        )

        # put all videos in db
        with DB() as db:
            db.insert_videos(all_videos)


def download_videos_to_be_dl():
    """
    Command dlsf
    Download videos in db with status TO_BE_DOWNLOADED OR (FAILED & fail_count < Config.)
    """
    with DB() as db:
        videos = db.query_videos(pending=True)

    vd = VideoDownloader()
    vd.downloads(videos, write_db=True, discover=False)


# --- --- --- --- --- DISCOVER --- --- --- --- ---
def discover(tag: str, max_page: int):
    discover_creators(tag, max_page)
    dl_video_from_discover_creators()


def discover_creators(tag: str, max_page: int):
    # download posts with searched tags
    posts = fetch_posts_by_tag(tag, max_page)
    logging.info(f"Find {len(posts)} post")

    path = str(Config.cache_file(tag))
    posts_with_videos = parser.filter_posts_with_videos_from_json(path)
    logging.info(f"Find {len(posts_with_videos)} posts with videos")

    creators = parser.get_creators_from_posts(posts_with_videos)

    # save to csv
    file = os.path.join(Config.DISCOVER_DIR, "discover.csv")
    with open(file, "w") as f:
        for c in creators:
            line = f"{c.creator_id};{c.service};{c.domain};{'to_be_treated'}\n"
            f.write(line)


def dl_video_from_discover_creators():
    # load csv
    file = os.path.join(Config.DISCOVER_DIR, "discover.csv")
    with open(file, "r") as f:
        lines = f.readlines()

    creators = []
    for line in lines:
        line = line.replace("\n", "").strip().split(";")
        creators.append(
            Creator(creator_id=line[0], service=line[1], domain=line[2], status=line[3])
        )

    # get posts
    for creator in creators:
        response = requests.get(
            URL.get_creator_post_wo_param(creator), headers=URL.get_headers()
        )
        if response.status_code != 200:
            print(f"ERROR - Request {URL.get_creator_post_wo_param(creator)}")
        response_posts = response.json()
        posts = parser.filter_posts_with_videos_from_list(response_posts)
        print(f"{len(posts)} found")
        if len(posts) > 5:
            posts = posts[0:5]
            print("Limited posts to 5")

        for post in posts:
            urls = parser.extract_video_urls(post)
            url = URL.get_url_from_file(creator.domain, urls[0])
            filename = f"{post['user']}_{post['id']}.mp4"
            filepath = os.path.join(Config.DISCOVER_DIR, filename)
            ytdlp_subprocess(url, filepath)



# core/parser.py

import logging
from pathvalidate import sanitize_filename

from .models import Video, VideoStatus, Creator
from .file_io import load_json, load_txt, write_txt
from .config import Config
from rcdl.interface.ui import UI


COOMER_PAYSITES = ["onlyfans", "fansly", "candfans"]
KEMONO_PAYSITES = [
    "patreon",
    "fanbox",
    "fantia",
    "boosty",
    "gumroad",
    "subscribestar",
    "dlsite",
]


def get_domain(arg: str | dict | Video) -> str:
    """From a service get the domain (coomer or kemono)
    Input is either: service(str), post(dict), video(models.Video)
    """

    def _service(service: str) -> str:
        if service in COOMER_PAYSITES:
            return "coomer"
        if service in KEMONO_PAYSITES:
            return "kemono"
        logging.error(f"Service {service} not associated to any domain")
        return ""

    if isinstance(arg, dict):
        return _service(arg["service"])
    elif isinstance(arg, Video):
        return _service(arg.service)

    return _service(arg)


def get_title(post: dict) -> str:
    """Extract title from a post(dict)"""
    title = post["title"]
    if title == "":
        if "content" in post:
            title = post["content"]
        elif "substring" in post:
            title = post["substring"]
    return sanitize_filename(title)


def get_date(post: dict) -> str:
    """Extract date from a post(dict)"""
    if "published" in post:
        date = post["published"][0:10]
    elif "added" in post:
        date = post["added"][0:10]
    else:
        logging.error(f"Could not extract date from {post['id']}")
        date = "NA"
    return date


def get_part(post: dict, url: str) -> int:
    """
    For posts containing multiple video url. Each url is considered a part,
    so all videos from the same posts will simply have a different part number
    """
    urls = extract_video_urls(post)
    part = 0
    if len(urls) == 1:
        return 0

    for u in urls:
        if u == url:
            return part
        part += 1

    logging.error(
        f"Could not extract part number for post id {post['id']} with url {url}"
    )
    return -1


def get_filename(post: dict, url: str) -> str:
    title = get_title(post)
    date = get_date(post)
    part = get_part(post, url)
    file_title = f"{date}_{title}".replace("'", " ").replace('"', "")
    filename = f"{file_title}_p{part}.mp4"
    return filename


def convert_post_to_video(post: dict, url: str, discover=False) -> Video:
    part = get_part(post, url)
    title = get_title(post)
    date = get_date(post)
    filename = get_filename(post, url)

    if discover:
        filename = f"{post['user']}_{post['id']}.mp4"

    return Video(
        post_id=post["id"],
        creator_id=post["user"],
        service=post["service"],
        relative_path=filename,
        url=url,
        domain=get_domain(post),
        part=part,
        published=date,
        title=title,
        status=VideoStatus.NOT_DOWNLOADED,
        fail_count=0,
    )


def convert_posts_to_videos(posts: list[dict], discover: bool = False) -> list[Video]:
    videos = []
    for post in posts:
        urls = extract_video_urls(post)
        if not discover:
            for url in urls:
                videos.append(convert_post_to_video(post, url))
        else:
            if len(urls) == 0:
                continue
            videos.append(convert_post_to_video(post, urls[0], discover=discover))
    return videos


def extract_video_urls(post: dict) -> list:
    video_extensions = (".mp4", ".webm", ".mov", ".avi", ".mkv", ".flv", ".wmv", ".m4v")
    urls = set()

    # Check main file
    if post["file"]:
        if post["file"]["path"]:
            path = post["file"]["path"]
            if path.endswith(video_extensions):
                urls.add(f"{path}")

    if post["attachments"]:
        attachments = post["attachments"]
        for attachment in attachments:
            if attachment["path"]:
                if attachment["path"].endswith(video_extensions):
                    urls.add(f"{attachment['path']}")

    return list(urls)


def filter_posts_with_videos_from_list(data: list[dict]) -> list[dict]:
    """Return posts with video url from a json with a list of posts"""

    posts_with_videos = []
    for post in data:
        if len(extract_video_urls(post)) > 0:
            posts_with_videos.append(post)
    return posts_with_videos


def filter_posts_with_videos_from_json(path: str) -> list:
    """Return posts with video url from a json with a list of posts"""
    posts = load_json(path)

    posts_with_videos = []
    for post in posts:
        if len(extract_video_urls(post)) > 0:
            posts_with_videos.append(post)
    return posts_with_videos


def valid_service(service: str) -> bool:
    if service in COOMER_PAYSITES:
        return True
    if service in KEMONO_PAYSITES:
        return True
    return False


def get_creator_from_line(line: str) -> Creator | None:
    """
    Convert a line into a Creator model
    arg: line -> 'service/creator'
    This is the format of creators.txt
    """
    parts = line.split("/")
    if valid_service(parts[0].strip()):
        return Creator(
            creator_id=parts[1].strip(),
            service=parts[0].strip(),
            domain=get_domain(parts[0].strip()),
            status=None,
        )
    elif valid_service(parts[1].strip()):
        return Creator(
            creator_id=parts[0].strip(),
            service=parts[1].strip(),
            domain=get_domain(parts[1].strip()),
            status=None,
        )
    else:
        UI.error(
            f"Creator file not valid: {line} can not be interpreted. Format is: 'service/creator_id'"
        )
    return None


def get_creators() -> list[Creator]:
    """
    Load creators.txt and return a list of models.Creator
    """
    lines = load_txt(Config.CREATORS_FILE)
    creators = []
    for line in lines:
        creator = get_creator_from_line(line)
        if creator is None:
            continue
        creators.append(creator)
    if len(creators) < 1:
        UI.error(f"Could not find any creators. Check {Config.CREATORS_FILE}")
    return creators


def get_creators_from_posts(posts: list[dict]) -> list[Creator]:
    creators = list()
    seen = set()

    for post in posts:
        key = (post["user"], post["service"], "coomer")
        if key in seen:
            continue

        seen.add(key)
        creators.append(
            Creator(
                creator_id=post["user"],
                service=post["service"],
                domain="coomer",
                status="to_be_treated",
            )
        )
    return creators


def parse_creator_input(value: str) -> tuple[str | None, str]:
    value = value.strip()

    # url
    if "://" in value:
        parts = value.replace("https://", "").strip().split("/")
        logging.info(f"From {value} extracte service {parts[1]} and creator {parts[3]}")
        return parts[1], parts[3]  # service, creator_id

    # creators.txt format
    if "/" in value:
        c = get_creator_from_line(value)
        if c is not None:
            logging.info(
                f"From {value} extracte service {c.service} and creator {c.creator_id}"
            )
            return c.service, c.creator_id

    logging.info(f"From {value} extracte service None and creator {value}")
    return None, value


def append_creator(creator: Creator):
    line = f"{creator.service}/{creator.creator_id}"
    lines = load_txt(Config.CREATORS_FILE)

    if line in lines:
        return
    lines.append(line)
    write_txt(Config.CREATORS_FILE, line, mode="a")



# core/downloader_subprocess.py

import subprocess
import logging
from pathlib import Path
import os

from rcdl.core.models import Video
from rcdl.core.config import Config
from rcdl.interface.ui import UI


def ytdlp_subprocess(
    url: str,
    filepath: Path | str,
):
    """Call yt-dlp in a subprocess"""
    cmd = [
        "yt-dlp",
        "-q",
        "--progress",
        url,
        "-o",
        filepath,
        "--external-downloader",
        "aria2c",
    ]

    logging.info(f"CMD: {' '.join(cmd)}")

    result = subprocess.run(cmd, capture_output=True, text=True)
    if result.returncode != 0:
        logging.error(f"yt-dlp failed to dl vid: {result.stderr}")

    return result.returncode


def ffmpeg_concat_build_command(videos: list[Video]) -> dict:
    # parameters
    width: int = 1920
    height: int = 1080
    fps: int = 30
    preset: str = "veryfast"
    threads: int = 0  # 0 for max

    # output path
    v = videos[0]
    output_filename = f"tmp_{v.published}_{v.title}.mp4"
    output_path = os.path.join(Config.creator_folder(v.creator_id), output_filename)

    # build cmd
    cmd = ["ffmpeg", "-y", "-progress", "pipe:2", "-nostats"]

    # inputs
    for v in videos:
        input_path = os.path.join(Config.creator_folder(v.creator_id), v.relative_path)
        cmd.extend(["-i", input_path])

    # filter complex
    filter_lines = []
    for idx in range(len(videos)):
        filter_lines.append(
            f"[{idx}:v]"
            f"scale={width}:{height}:force_original_aspect_ratio=decrease,"
            f"pad={width}:{height}:(ow-iw)/2:(oh-ih)/2,"
            f"fps={fps},setsar=1"
            f"[v{idx}]"
        )

    # concat inputs
    concat = []
    for idx in range(len(videos)):
        concat.append(f"[v{idx}][{idx}:a]")

    filter_lines.append(f"{''.join(concat)}concat=n={len(videos)}:v=1:a=1[outv][outa]")
    filter_complex = ";".join(filter_lines)

    cmd.extend(
        [
            "-filter_complex",
            filter_complex,
            "-map",
            "[outv]",
            "-map",
            "[outa]",
            "-c:v",
            "libx264",
            "-preset",
            preset,
            "-threads",
            str(threads),
            "-c:a",
            "aac",
            "-f",
            "mp4",
            output_path,
        ]
    )

    return {"cmd": cmd, "output_path": output_path}


def get_duration(path: str) -> int:
    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-select_streams",
        "v:0",
        "-show_entries",
        "format=duration",
        "-of",
        "default=noprint_wrappers=1:nokey=1",
        path,
    ]

    result = subprocess.run(cmd, capture_output=True, text=True, check=True)
    return int(float(result.stdout.strip()) * 1000)


def get_total_duration(videos: list[Video]) -> int:
    duration = 0
    for v in videos:
        path = os.path.join(Config.creator_folder(v.creator_id), v.relative_path)
        duration += get_duration(path)
    return duration


def ffmpeg_concat(videos: list[Video]):
    command_builder = ffmpeg_concat_build_command(videos)
    cmd = command_builder["cmd"]
    output_path = command_builder["output_path"]

    logging.info(f"CMD: {' '.join(cmd)}")

    # run cmd
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.PIPE,
        text=True,
        bufsize=1,
    )

    assert process.stderr is not None
    total_duration = get_total_duration(videos)
    progress, task = UI.concat_progress(total=total_duration)
    last_progress = 0
    UI.set_current_concat_progress(f"{videos[0].relative_path}", output_path)

    for line in process.stderr:
        line = line.strip()
        if not line:
            continue

        progres_key = "out_time_ms"
        if line.startswith(progres_key):
            current_progress_str = line.replace(f"{progres_key}=", "").strip()
            try:
                current_progress_us = int(current_progress_str)
                current_progress_ms = current_progress_us // 1000
                delta = current_progress_ms - last_progress
                progress.advance(task, advance=delta)
                last_progress = current_progress_ms
            except Exception:
                pass

    UI.close_concat_progress()

    if process.returncode != 0:
        UI.error("Failed to concat videos")
        return process.returncode

    os.replace(output_path, output_path[:-4])
    UI.info(f"Rename {output_path} -> {output_path}")
    return 0



# scripts/upload_pypi.py

import subprocess
import sys
import os
import requests
from packaging.version import parse as parse_version
import tomllib

# config
PYPROJECT_FILE = "pyproject.toml"
PYPI_PACKAGE_NAME = "rcdl"  # PyPI package name

# read local version
with open(PYPROJECT_FILE, "rb") as f:
    data = tomllib.load(f)

local_version = data["project"]["version"]
print(f"Local version: {local_version}")

# check remote latest version
response = requests.get(f"https://pypi.org/pypi/{PYPI_PACKAGE_NAME}/json")
if response.status_code == 200:
    latest_version = response.json()["info"]["version"]
    print(f"Latest PyPI version: {latest_version}")
else:
    latest_version = None
    print("ERROR: Package not found on PyPI")

# chec version number
if latest_version and parse_version(local_version) <= parse_version(latest_version):
    print("Error: Local version is not higher than PyPI version.")
    sys.exit(1)

# build with flit
print("Building package...")
subprocess.run([sys.executable, "-m", "flit", "build"], check=True)

# upload to pypi
print("Uploading to PyPI...")
if not os.path.exists("api_key.txt"):
    print(
        "ERROR - you have to create an api_key.txt file in your root directory containing only your pypi api key"
    )
    quit()
with open("api_key.txt", "r") as f:
    api_key = f.read().strip()
if api_key == "":
    print("ERROR - api_key.txt is empty")

subprocess.run(
    [
        sys.executable,
        "-m",
        "twine",
        "upload",
        "-u",
        "__token__",
        "-p",
        api_key,
        "dist/*",
    ],
    check=True,
)

print("PYPI Upload complete!")

# find wheel file in dist/
dist_files = [f for f in os.listdir("dist") if f.endswith(".whl")]
if not dist_files:
    raise FileNotFoundError("No .whl file found in dist/")

whl_file_path = os.path.join("dist", dist_files[0])
if local_version not in whl_file_path:
    raise FileNotFoundError("Version of .whl does not match upload.")

# create github release
tag = f"v{local_version}"
print(f"Creating github release {tag}")


subprocess.run(
    [
        "gh",
        "release",
        "create",
        tag,
        "--title",
        f"Release {tag}",
        "--notes",
        "",
        whl_file_path,
    ],
    check=True,
)

print("GitHub release created!")
print("--END--")



# core/db_queries.py

"""
Hold all the SQL commands strings
"""

CREATE_VIDEOS_TABLE = """
CREATE TABLE IF NOT EXISTS videos (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    post_id TEXT,
    creator_id TEXT,
    service TEXT,
    domain TEXT,
    relative_path TEXT,
    url TEXT,
    part TEXT,
    status TEXT DEFAULT 'not_downloaded',
    fail_count INTEGER DEFAULT 0,
    published TEXT,
    title TEXT,
    substring TEXT,
    downloaded_at TEXT,
    file_size REAL,
    UNIQUE (service, url)
)
"""

CREATE_SCHEMA_VERSION_TABLE = """
CREATE TABLE IF NOT EXISTS schema_version (
    version INTEGER NOT NULL
)
"""

INSERT_VIDEO_UPSERT = """
INSERT INTO videos (
    post_id, creator_id, service, domain, relative_path, url, part,
    status, fail_count, published, title, substring,
    downloaded_at, file_size
)
VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
ON CONFLICT(service, url) DO UPDATE SET
    status = excluded.status,
    fail_count = excluded.fail_count,
    relative_path = excluded.relative_path,
    downloaded_at = excluded.downloaded_at,
    file_size = excluded.file_size
"""

INSERT_IGNORE_VIDEO_UPSERT = """
INSERT OR IGNORE INTO videos (
    post_id, creator_id, service, domain, relative_path, url, part, 
    status, fail_count, published, title, substring, 
    downloaded_at, file_size
)
VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
"""

CREATE_IDX_VIDEOS_STATUS = (
    "CREATE INDEX IF NOT EXISTS idx_videos_status ON videos(status)"
)
CREATE_IDX_VIDEOS_CREATOR = (
    "CREATE INDEX IF NOT EXISTS idx_videos_creator ON videos(creator_id)"
)
CREATE_IDX_VIDEOS_FAIL_COUNT = (
    "CREATE INDEX IF NOT EXISTS idx_videos_fail_count ON videos(fail_count)"
)



# core/db.py

"""Handle SQL DB and DB Parsing"""

import sqlite3
import logging
from typing import Iterable


import rcdl.core.db_queries as queries
from .config import Config
from .models import Video, VideoStatus


class DB:
    def __init__(self):
        self.conn = sqlite3.connect(Config.DB_PATH)
        self.conn.row_factory = sqlite3.Row
        self.conn.execute("PRAGMA foreign_keys = ON")  # check input into db
        self.conn.execute("PRAGMA journal_mode = WAL")  #
        self.conn.execute("PRAGMA synchronous = NORMAL")  # faster write speed

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def init_table(self):
        # init table for videos to DL
        self.conn.execute(queries.CREATE_VIDEOS_TABLE)
        self.conn.execute(queries.CREATE_IDX_VIDEOS_STATUS)
        self.conn.execute(queries.CREATE_IDX_VIDEOS_CREATOR)
        self.conn.execute(queries.CREATE_IDX_VIDEOS_FAIL_COUNT)

        # init table for version schem, easy migration if necessary
        self.conn.execute(queries.CREATE_SCHEMA_VERSION_TABLE)

        # Initialize version if empty
        cur = self.conn.cursor()
        cur.execute("SELECT COUNT(*) as cnt FROM schema_version")
        row = cur.fetchone()
        if row["cnt"] == 0:
            cur.execute("INSERT INTO schema_version (version) VALUES (?)", (1,))

        self.conn.commit()

    def get_schema_version(self) -> int:
        # get current db version -> for future update to db
        cur = self.conn.cursor()
        cur.execute("SELECT version FROM schema_version")
        row = cur.fetchone()
        return row["version"] if row else 0

    def set_schema_version(self, version: int):
        # set current db version
        self.conn.execute("UPDATE schema_version SET version = ?", (version,))
        self.conn.commit()

    def _row_to_video(self, row: sqlite3.Row) -> Video:
        # helper function to retunr Video model from a sql result
        return Video(
            post_id=row["post_id"],
            creator_id=row["creator_id"],
            service=row["service"],
            domain=row["domain"],
            relative_path=row["relative_path"],
            url=row["url"],
            part=row["part"],
            status=VideoStatus(row["status"]),
            fail_count=row["fail_count"],
            published=row["published"],
            title=row["title"],
            substring=row["substring"],
            downloaded_at=row["downloaded_at"],
            file_size=row["file_size"],
        )

    def query_videos(
        self,
        *,
        status: VideoStatus | Iterable[VideoStatus] | None = None,
        creator_id: str | None = None,
        post_id: str | None = None,
        max_fail_count: int | None = None,
        min_part_number: int | None = None,
        pending: bool = False,
    ) -> list[Video]:
        """
        Query the DB.
        Parametes are 'AND' so if video status and creator_id are sepcified,
        it will look for a match between the two

        pending:
        No parameters are taken into account if pending True.
        Look for videos with: NOT DOWNLOADED status OR (FAILED & fail_count < Config.max_fail_count)
        """
        sql = "SELECT * FROM videos"
        conditions = []
        params = []

        if pending:
            sql += " WHERE status = ? OR (status = ? AND fail_count < ?)"
            params.extend(
                [
                    VideoStatus.NOT_DOWNLOADED.value,
                    VideoStatus.FAILED.value,
                    max_fail_count or Config.MAX_FAIL_COUNT,
                ]
            )
        else:
            if status is not None:
                if isinstance(status, VideoStatus):
                    conditions.append("status = ?")
                    params.append(status.value)
                else:
                    statuses = list(status)
                    placeholders = ",".join("?" for _ in statuses)
                    conditions.append(f"status IN ({placeholders})")
                    params.extend(s.value for s in statuses)

            if creator_id is not None:
                conditions.append("creator_id = ?")
                params.append(creator_id)

            if post_id is not None:
                conditions.append("post_id = ?")
                params.append(post_id)

            if max_fail_count is not None:
                conditions.append("fail_count < ?")
                params.append(max_fail_count)

            if min_part_number is not None:
                conditions.append("part > ?")
                params.append(min_part_number)

            if conditions:
                sql += " WHERE " + " AND ".join(conditions)

        logging.debug(f"SQL CMD: {sql} with params: {params}")
        cur = self.conn.cursor()
        cur.execute(sql, params)
        rows = cur.fetchall()
        if Config.DEBUG:
            logging.debug(f"DB query returned {len(rows)} result")

        return [self._row_to_video(r) for r in rows]

    def get_db_videos_info(self):
        """Return number of videos per status
        return info: dict {
            "not_downloaded": int,
            "failed": int,
            etc...
        }
        """
        info = {}
        for status in VideoStatus:
            vids = self.query_videos(status=status)
            info[status.value] = len(vids)
        return info

    def set_status(
        self, video: Video, status: VideoStatus, *, fail_count: int | None = None
    ):
        """Set video status to specified status"""
        video.status = status
        if fail_count is not None:
            video.fail_count = fail_count
        self._upsert_video(video)

    def insert_videos(self, videos: list[Video]):
        """
        Insert a video if not already present. Else ignore.
        Does not modify any values
        """
        if not videos:
            return

        rows = []
        for video in videos:
            rows.append(
                (
                    video.post_id,
                    video.creator_id,
                    video.service,
                    video.domain,
                    video.relative_path,
                    video.url,
                    video.part,
                    VideoStatus.NOT_DOWNLOADED.value,
                    0,
                    video.published,
                    video.title,
                    video.substring,
                    None,
                    None,
                )
            )

        self.conn.executemany(queries.INSERT_IGNORE_VIDEO_UPSERT, rows)
        self.conn.commit()

    def _upsert_video(self, video: Video):
        """Upsert a video.
        If video already in DB, update specifics fields:
             status, fail_count, relative path, file_size, downloaded_at
        """
        if video.status is None:
            video.status = VideoStatus.NOT_DOWNLOADED

        self.conn.execute(
            queries.INSERT_VIDEO_UPSERT,
            (
                video.post_id,
                video.creator_id,
                video.service,
                video.domain,
                video.relative_path,
                video.url,
                video.part,
                video.status.value,
                video.fail_count,
                video.published,
                video.title,
                video.substring,
                video.downloaded_at,
                video.file_size,
            ),
        )
        self.conn.commit()

    def close(self):
        self.conn.close()



# interface/cli.py

import logging

import click

from rcdl.core import downloader as dl
from rcdl.core.config import Config
from rcdl.core.parser import (
    get_creators,
    get_creator_from_line,
    parse_creator_input,
    append_creator,
)
from rcdl.core.db import DB
from .ui import UI
from rcdl.core.fuse import fuse_videos


from rcdl import __version__


@click.command(help="Refresh video to be downloaded")
def refresh():
    """Refresh database with creators videos

    - get all creators from creators.txt
    - for each creators find all videos and put them in the database
    No download is done in this command
    """

    UI.info("Welcome to RCDL refresh")
    dl.refresh_creators_videos()

    with DB() as db:
        info = db.get_db_videos_info()

    UI.db_videos_status_table(info)


@click.command(help="Download all videos from all creator")
def dlsf():
    """Download video based on DB information

    - read databse
    - for each video NOT_DOWNLOADED or FAILED & fail_count < settings, dl video
    """
    UI.info("Welcome to RCDL dlsf")
    dl.download_videos_to_be_dl()


@click.command("fuse", help="Fuse part video into one")
def fuse():
    """Fuse videos"""
    UI.info("fuse")
    fuse_videos()


@click.command(help="Discover videos/creators with tags")
@click.option("--tag", required=True, type=str, help="Tag to search for")
@click.option(
    "--max-page", default=10, type=int, help="Maximum number of pages to fetch"
)
def discover(tag, max_page):
    """Discover new creators/videos
    currently WIP. Do not use in prod"""
    msg = f"[cdl] discover with tag={tag} max_page={max_page}"
    click.echo(msg)
    logging.info(msg)
    dl.discover(tag, max_page)


@click.command("add", help="Add a creator")
@click.argument("creator_input")
def add_creator(creator_input):
    """Add a creator (URL or str) to creators.txt"""
    service, creator_id = parse_creator_input(creator_input)
    line = f"{service}/{creator_id}"
    creator = get_creator_from_line(line)
    if creator is not None:
        append_creator(creator)
        UI.info(f"Added {line} to creators.txt")
    else:
        UI.warning("Could not extract creator from input. Please check input is valid")


@click.command("remove", help="Remove a creator")
@click.argument("creator_input")
def remove_creator(creator_input):
    """Remove a creator (excat line) from creators.txt"""
    _service, creator_id = parse_creator_input(str(creator_input))

    creators = get_creators()
    all_creators = []
    matched_creator = None
    for creator in creators:
        if creator.creator_id == creator_id:
            matched_creator = creator
            continue
        all_creators.append(creator)

    if matched_creator is None:
        UI.error(f"Could not find creator from {creator_input}")
        return
    else:
        open(Config.CREATORS_FILE, "w").close()
        for c in all_creators:
            append_creator(c)
        UI.info(
            f"Removed creator {matched_creator.creator_id}@({matched_creator.service})"
        )


@click.command("list", help="List all creators")
def list_creators():
    creators = get_creators()
    UI.table_creators(creators)


# --- CLI GROUP ---
@click.group()
@click.option("--debug", is_flag=True)
@click.option("--dry-run", is_flag=True)
@click.version_option(version=__version__, prog_name=Config.APP_NAME)
def cli(debug, dry_run):
    Config.set_debug(debug)
    Config.set_dry_run(dry_run)


cli.add_command(dlsf)
cli.add_command(discover)
cli.add_command(refresh)
cli.add_command(add_creator)
cli.add_command(remove_creator)
cli.add_command(list_creators)
cli.add_command(fuse)
